
' *****************************************************************************
' *                                                                           *
' *     Ethernet_Driver.spin                                                  *
' *                                                                           *
' *****************************************************************************

' Known Bugs, Issues and Features
' ===============================
'
' No NLP timeout
' Requires 80MHz CLKFREQ
' No check on CLKFREQ
' Should determine FRQA and PLLDIV for CTRA from CLKFREQ
' Should determine timeout16uS based on CLKFREQ

CON

  ETH_BYTES_MAX     = 1500                              ' Max size of Ethernet Packet in total
  ETH_BYTES_MIN     = 64

  ETH_BITS_PER_LONG = 16                                ' 16 = Manchester Encoded / 32 = Debugging
  
  ETH_BITS          = ETH_BYTES_MAX * 8
  ETH_LONGS         = ETH_BITS / ETH_BITS_PER_LONG
  
  IP_LEN_MSB        = 24
  IP_LEN_LSB        = 25

VAR

  long bitArray[ ETH_LONGS + 1 ]
  
  word bitArrayIndex
  long bitMask

' *****************************************************************************
' *                                                                           *
' *     Start the Assembler Cog running                                       *
' *                                                                           *
' *****************************************************************************

PUB Start( ethernetBasePin ) | vGroup, vPins, pShift

  ethernetBasePin &= $1E                                  ' D- = 0..30, even numbered ( D+ = D- plus 1 )

  vGroup := ethernetBasePin / 8                           ' vGroup = 0..7, 8..15, 16..23, 24..31
  vPins  := %11 << ( ethernetBasePin & 7 )                ' vPins = 00000011, 00001100, 00110000, 11000000
  pShift := ethernetBasePin
  
  ' %-------- ---XXXXX -------- --------    pShift
  ' %-------- -------- ----XXX- --------    vGroup
  ' %-------- -------- -------- XXXXXXXX    vPins
  
  bitArray[0] := ( pShift << 16 ) | ( vGroup << 9 ) | vPins

  cognew( @BitBangedEthernetDriver, @bitArray )

' *****************************************************************************
' *                                                                           *
' *     Take a byte packet, convert to bits and let the Cog stream it out     *
' *                                                                           *
' *****************************************************************************

PUB CreateAndSendBitStream( pointerToByteArray ) | endNum, i, longsToSend

  bitArrayIndex := 0                                    ' First bit
  bitMask := 0
  
  endNum := ( ( byte[ pointerToByteArray+IP_LEN_MSB ] << 8 ) | byte[ pointerToByteArray+IP_LEN_LSB ] ) + 25

  repeat i from 0 to endNum                             ' Convert bytes
    SetBitsFromByte( byte[ pointerToByteArray+i ] )

  longsToSend := ( endNum * 8 ) / ETH_BITS_PER_LONG     ' How many longs to send
  if (( longsToSend * ETH_BITS_PER_LONG ) / 8 ) <> endNum
    longsToSend++

  repeat until bitArray[0] == 0                         ' Wait until last sent
    
  bitArray[0] := longsToSend                            ' Send a new bit stream

PRI SetBitsFromByte( n )

' Converting bytes to bits is straight forward but it may not be immediately clear as to how it is done
' and how it works in total ...
'
' Each bit ( sent lsb first from the byte ) is Manchester Encoded into two bits. According to the Ethernet
' Blue-Book,a 0-bit is sent 0 first / 1 second, a 1-bit sent as 1 first / 0 second. Because the longs we
' are building will be sent lsb first, the first encoded bit is placed to the right of the second ...
'
' A byte to send : ------0-1
'                        | |
'                        | `---------------------------.
'                        `-------------------------.   |
'                                                 _|  _|
' A long to send : -------- -------- -------- ----10--01
'
' As can be seen, we can hold 16 MTU bits in each long.

' The output is also bi-phase and we use the colour registers to decide which output pins will be set and
' cleared. Each bit in the long will choose one of two colours, each colour will set one of two pins with
' the other being cleared. The two colours are opposites of each other.
'
' A byte to send : ------0-1
'                        | |
'                        | `---------------------------.
'                        `-------------------------.   |
'                                                 _|  _|
' A long to send : -------- -------- -------- ----10--01 
'                                               
' Output Pin D+  : -------- -------- -------- ----10--01  
' Output Pin D-  : -------- -------- -------- ----01--10

  if bitMask == 0                                       ' Setup for first bit
    bitArray[++bitArrayIndex] := 0
    bitMask := 1
    
  if ETH_BITS_PER_LONG == 16                            ' Manchester Encoded 
  
    repeat 8
      if n & 1
        bitArray[ bitArrayIndex ] |= ( bitMask      )   ' 1 -> 01
      else
        bitArray[ bitArrayIndex ] |= ( bitMask << 1 )   ' 0 -> 10
      bitMask <<= 2
      n >>= 1

  else                                                  ' Bit-for-Bit ( Debugging use )
  
    repeat 8
      if n & 1
        bitArray[ bitArrayIndex ] |= bitMask
      bitMask <<= 1
      n >>= 1
      
PUB GetPointerToBitArray

  result := @bitArray[0]

DAT

' *****************************************************************************
' *                                                                           *
' *     Cog Assembler Code                                                    *
' *                                                                           *
' *****************************************************************************

              org       0
              
BitBangedEthernetDriver

              mov       streamPtr,PAR                   ' Point to bit array
              rdlong    bits,streamPtr                  ' Determine configuration to use

              ' %-------- ---XXXXX -------- --------    pShift
              ' %-------- -------- ----XXX- --------    vGroup
              ' %-------- -------- -------- XXXXXXXX    vPins
              
              movs      vcfgInit,bits                   ' Set VCFG VPins
              
              shr       bits,#9                         ' Set VCFG VGroup
              movd      vcfgInit,bits

              shr       bits,#7                         ' Initialise I/O pins    
              mov       count,bits                      
              mov       bits,#%11
              shl       bits,count
              
              mov       OUTA,#0                         
              mov       DIRA,bits                       ' Make D- and D+ pins outputs
              
              mov       FRQA,frqaInit                   ' Initialise Video Hardware
              mov       CTRA,ctraInit
              mov       VCFG,vcfgInit
              mov       VSCL,vsclInit

ReadyForNextStream

              mov       streamPtr,PAR                   ' Point to bit array
              wrlong    zero,streamPtr                  ' Tell sender we're ready to accept a stream

StartTimeout

              mov       timeout,CNT                     ' Set timeout for 16mS +/- 8mS
              add       timeout,timeout16mS

ShortPauseAfterAnythingSent

              mov       count,#200                      ' Delay at least 9.6uS ( inter-frame gap )
Delay10uS     djnz      count,#Delay10uS                ' 200 x 50nS per instruction = 10000nS = 10uS
              
WaitForStream

              rdlong    count,streamPtr         WZ      ' Wait until stream provided ( non-zero count )
        IF_Z  jmp       #CheckForTimeout                ' Not ready yet

OutputBitStream
        
              add       streamPtr,#4                    ' Point to long to output bitArray[1..count]
              rdlong    bits,streamPtr
        
              waitvid   colours,bits
              
              djnz      count,#OutputBitStream

              jmp       #ReadyForNextStream
              
CheckForTimeout

              ' If (timeout - CNT ) <= 0 Then TimedOut
              
              sub       timeout,CNT         NR, WC, WZ  ' Has timeout occurred ?
        IF_A  jmp       #WaitForStream                  ' No

TimedOut

              waitvid colours,#%1111                    ' Send NLP - 100nS pulse = 4 x 25nS (20MHz)
              
              jmp       #StartTimeout

' *****************************************************************************
' *                                                                           *
' *     Predefined Constants                                                  *
' *                                                                           *
' *****************************************************************************

zero          long      0                               ' Constant value zero

frqaInit      long      $8000_0000 >> 4                 ' 80MHz / 16 x PLLx16 = 80MHz

ctraInit      long      ( %00001 << 26 ) | ( %101 << 23 )

              ' %-00001-- -------- -------- --------    CtrMode      000001 = Video PLL
              ' %------10 1------- -------- --------    PllDiv       101    = Divide by 4 = 20MHz
              ' %-------- -------- -XXXXXX- --------    BPin         XXXXXX = Not Used
              ' %-------- -------- -------- --XXXXXX    APin         XXXXXX = Not Used

vcfgInit      long      %00100000_00000000_00000000_00000000

              ' %-01----- -------- -------- --------    VMode        01     = VGA
              ' %---0---- -------- -------- --------    CMode        0      = Two Colour Mode
              ' %----00-- -------- -------- --------    Chroma1/0    00     = No Chroma
              ' %------00 0------- -------- --------    Aural Sub    000    = No Aural Sub Cog
              ' %-------- -------- ----XXX- --------    VGroup       XXX    = Set At Run-Time
              ' %-------- -------- -------- XXXXXXXX    VPins        XXX    = Set At Run-Time
              '                 DD DDDDDDDS SSSSSSSS

vsclInit      long      ( 1 << 12 ) | 32

              ' %-------- ----0000 0001---- --------    PixelClocks  1      = 1 bit per 20MHz
              ' %-------- -------- ----0000 00100000    FrameClocks  32     = 32 bits per long

colours       long      $00_00_AA_55                    ' D+ to left, D- to right of each two-bit pair

              ' %-------- -------- -------- 01010101    Colour 0     Bit Clear ( D- = 1, D+ = 0 )
              ' %-------- -------- 10101010 --------    Colour 1     Bit Set   ( D- = 0, D+ = 1 )
              ' %-------- XXXXXXXX -------- --------    Colour 2     Not Used
              ' %XXXXXXXX -------- -------- --------    Colour 3     Not Used

timeout16mS   long      80_000_000 / 1000 * 16        ' 16mS at 80MHz 

' *****************************************************************************
' *                                                                           *
' *     Run-Time Variables                                                    *
' *                                                                           *
' *****************************************************************************

streamPtr     res       1
bits          res       1
count         res       1
timeout       res       1
tmp           res       1

              fit       $1F0
              
